﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  public class M5RuleSetModel : RegressionModel, IM5MetaModel {
    private const string NoRulesResultName = "Number of Rules";
    private const string CoveredInstancesResultName = "Covered Instances";

    #region Properties
    [Storable]
    internal List<M5RuleModel> Rules { get; private set; }
    #endregion

    #region HLConstructors & Cloning
    [StorableConstructor]
    protected M5RuleSetModel(bool deserializing) : base(deserializing) { }
    protected M5RuleSetModel(M5RuleSetModel original, Cloner cloner) : base(original, cloner) {
      if (original.Rules != null) Rules = original.Rules.Select(cloner.Clone).ToList();
    }
    protected M5RuleSetModel(string targetVariable) : base(targetVariable) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new M5RuleSetModel(this, cloner);
    }
    #endregion

    internal static M5RuleSetModel CreateRuleModel(string targetAttr, M5CreationParameters m5CreationParams) {
      return m5CreationParams.LeafType is ILeafType<IConfidenceRegressionModel> ? new ConfidenceM5RuleSetModel(targetAttr) : new M5RuleSetModel(targetAttr);
    }

    #region RegressionModel
    public override IEnumerable<string> VariablesUsedForPrediction {
      get {
        var f = Rules.FirstOrDefault();
        return f != null ? (f.VariablesUsedForPrediction ?? new List<string>()) : new List<string>();
      }
    }
    public override IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      if (Rules == null) throw new NotSupportedException("The classifier has not been built yet");
      return rows.Select(row => GetEstimatedValue(dataset, row));
    }
    public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, problemData);
    }
    #endregion

    #region IM5Component
    void IM5MetaModel.BuildClassifier(IReadOnlyList<int> trainingRows, IReadOnlyList<int> holdoutRows, M5CreationParameters m5CreationParams, CancellationToken cancellation) {
      Rules = new List<M5RuleModel>();
      var tempTraining = trainingRows;
      var tempHoldout = holdoutRows;
      do {
        var tempRule = M5RuleModel.CreateRuleModel(m5CreationParams.TargetVariable, m5CreationParams);
        cancellation.ThrowIfCancellationRequested();

        if (!m5CreationParams.Results.ContainsKey(NoRulesResultName)) m5CreationParams.Results.Add(new Result(NoRulesResultName, new IntValue(0)));
        if (!m5CreationParams.Results.ContainsKey(CoveredInstancesResultName)) m5CreationParams.Results.Add(new Result(CoveredInstancesResultName, new IntValue(0)));

        var t1 = tempTraining.Count;
        tempRule.BuildClassifier(tempTraining, tempHoldout, m5CreationParams, cancellation);
        tempTraining = tempTraining.Where(i => !tempRule.Covers(m5CreationParams.Data, i)).ToArray();
        tempHoldout = tempHoldout.Where(i => !tempRule.Covers(m5CreationParams.Data, i)).ToArray();
        Rules.Add(tempRule);
        ((IntValue) m5CreationParams.Results[NoRulesResultName].Value).Value++;
        ((IntValue) m5CreationParams.Results[CoveredInstancesResultName].Value).Value += t1 - tempTraining.Count;
      }
      while (tempTraining.Count > 0);
    }

    void IM5MetaModel.UpdateModel(IReadOnlyList<int> rows, M5UpdateParameters m5UpdateParameters, CancellationToken cancellation) {
      foreach (var rule in Rules) rule.UpdateModel(rows, m5UpdateParameters, cancellation);
    }
    #endregion

    #region Helpers
    private double GetEstimatedValue(IDataset dataset, int row) {
      foreach (var rule in Rules) {
        var prediction = rule.GetEstimatedValues(dataset, row.ToEnumerable()).Single();
        if (rule.Covers(dataset, row)) return prediction;
      }
      throw new ArgumentException("Instance is not covered by any rule");
    }
    #endregion

    [StorableClass]
    private class ConfidenceM5RuleSetModel : M5RuleSetModel, IConfidenceRegressionModel {
      #region HLConstructors & Cloning
      [StorableConstructor]
      protected ConfidenceM5RuleSetModel(bool deserializing) : base(deserializing) { }
      private ConfidenceM5RuleSetModel(ConfidenceM5RuleSetModel original, Cloner cloner) : base(original, cloner) { }
      public ConfidenceM5RuleSetModel(string targetVariable) : base(targetVariable) { }
      public override IDeepCloneable Clone(Cloner cloner) {
        return new ConfidenceM5RuleSetModel(this, cloner);
      }
      #endregion

      #region IConfidenceRegressionModel
      public IEnumerable<double> GetEstimatedVariances(IDataset dataset, IEnumerable<int> rows) {
        if (Rules == null) throw new NotSupportedException("The classifier has not been built yet");
        return rows.Select(row => GetEstimatedVariance(dataset, row));
      }
      public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
        return new ConfidenceRegressionSolution(this, problemData);
      }
      private double GetEstimatedVariance(IDataset dataset, int row) {
        foreach (var rule in Rules) {
          var prediction = ((IConfidenceRegressionModel) rule).GetEstimatedVariances(dataset, row.ToEnumerable()).Single();
          if (rule.Covers(dataset, row)) return prediction;
        }
        throw new ArgumentException("Instance is not covered by any rule");
      }
      #endregion
    }
  }
}